-- Class: Fire effect
-- SDK: Corona - 09/12/11
-- Author: Andrew Burch
-- Date: 15/04/2012
-- Site: http://www.newhighscore.net
-- Contact: andrew.burch@newhighscore.net
-- Notes: Animated fire demo effect
--

local core = require('core.core')
local tablex = require('core.tablex')

local clamp1 = core.clamp1
local floor = math.floor
local mathmax = math.max
local rng = math.random
local _ipairs = ipairs

local Fire = {}
local Fire_mt = {__index = Fire}

function Fire.new(params)
	local displayGroup = display.newGroup()
	
	local fire = {
		displayGroup = displayGroup,
	}
	
	return setmetatable(fire, Fire_mt)
end

function Fire:initialise(params)
	local displayGroup = self.displayGroup

	local paletteSize = params.paletteSize
	local cellWidth = params.cellWidth
	local cellHeight = params.cellHeight	
	local cellsPerRow = params.cellsPerRow
	local cellRowCount = params.cellRowCount	
	local top = params.top
	local initialLeft = params.left
	local sideBuffer = params.sideBuffer
	
	local displayCells = {}
	
	-- create display objects
	for y = 0, cellRowCount do
		local rowCells = {}
		local left = initialLeft
		for x = 0, cellsPerRow - 1 do
			local cell = display.newRect(displayGroup, left, top, cellWidth, cellHeight)
			local initAsBlank = (x < sideBuffer) or (x > cellsPerRow - cellsPerRow)
			local initColourIndex = initAsBlank and 0 or rng(0, paletteSize - 1)
			rowCells[x] = { 
					object = cell,
					colourIndex = initColourIndex,
				}
			
			left = left + cellWidth
		end
		
		displayCells[y] = rowCells
		top = top + cellHeight
	end
		
	-- build colour table
	local clearColour = params.clearColour
	local colourChart = params.colourChart
	local colourCount = tablex.icount(colourChart) - 1
	local shadesPerColour = paletteSize / colourCount
	local palette = {}
	local pindex = 0
	
	for i = 1, colourCount do
		local previousColour = colourChart[i - 1] or clearColour
		local currentColour = colourChart[i]
		local rstep = (currentColour.r - previousColour.r) / shadesPerColour
		local gstep = (currentColour.g - previousColour.g) / shadesPerColour
		local bstep = (currentColour.b - previousColour.b) / shadesPerColour
		
		for j = 0, shadesPerColour - 1 do
			local red = previousColour.r + (rstep * j)
			local green = previousColour.g + (gstep * j)
			local blue = previousColour.b + (bstep * j)
		
			palette[pindex] = {
						r = clamp1(red, 0, 255),
						g = clamp1(green, 0, 255),
						b = clamp1(blue, 0, 255),
						a = currentColour.a,
					}
							
			pindex = pindex + 1
		end
	end
	
	self.palette = palette
	self.newPaletteMin = params.newColourMin
	self.newPaletteMax = params.newColourMax
	self.displayCells = displayCells
	self.cellsPerRow = cellsPerRow
	self.cellRowCount = cellRowCount
	self.sideBuffer = sideBuffer
end

function Fire:update(dt, time, params)
	local cellsPerRow = self.cellsPerRow
	local cellRowCount = self.cellRowCount
	local displayCells = self.displayCells
	local palette = self.palette
	local fadeAccelerate = self.fadeAccelerate
	
	local filter = params.filter
	local fadeAccelerate = params.fadeAccelerate
	
	local newPaletteMin = self.newPaletteMin
	local newPaletteMax = self.newPaletteMax
	local sideBuffer = self.sideBuffer
	
	-- allocate new random base
	for x = sideBuffer, cellsPerRow - sideBuffer do
		local cellInfo = displayCells[cellRowCount][x]
		local colourIndex = rng(newPaletteMin, newPaletteMax)
		local colour = palette[colourIndex]		
		local object = cellInfo.object
		
		object:setFillColor(colour.r, colour.g, colour.b)
		cellInfo.colourIndex = colourIndex
	end
	
	-- apply filter
	for y = 0, cellRowCount - 1 do
		for x = 0, cellsPerRow - 1 do
			local total = 0
			local divider = 0
			
			for i,v in _ipairs(filter) do
				local row = displayCells[y + v.y]
				local cell = row and row[x + v.x]
				
				if cell then
					total = total + cell.colourIndex
					divider = divider + 1
				end
			end

			total = mathmax(0, total - fadeAccelerate)
			
			local newColourIndex = floor(total / divider)
			local colour = palette[newColourIndex]
			local cell = displayCells[y][x]
			local object = cell.object
			object:setFillColor(colour.r, colour.g, colour.b)
			cell.colourIndex = newColourIndex
		end
	end
end

return Fire
